# Generated by Django 2.0.13 on 2019-02-28 16:55

import django.contrib.postgres.fields.jsonb
from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('geostore', '0017_auto_20181114_0806'),
    ]

    operations = [
        migrations.AlterField(
            model_name='feature',
            name='properties',
            field=django.contrib.postgres.fields.jsonb.JSONField(blank=True, default=dict),
        ),
    ]
